package body RCP.Control is

   protected body Controller is
      entry Allocate (Res    :    out Resource_T;
                      Amount :        Request_T) when Free > 0 is
         --  we accept the request if we have spare resources to offer
      begin
         --  but we satisfy it only if have enough for it
         if Amount <= Free then
            --  the lucky user has got all the items
            --+ that it wanted the first time round!
            Res.Granted := Amount;
            Free := Free - Amount;
         else
            --  otherwise we transfer the request to an internal queue
            --+ after keeping record of the smallest request that
            --+ we were not able to satisfy
            --+ (this value will help us update the guard to
            --+  the private channel Assign)
            if Amount < Min_Request then
               Min_Request := Amount;
            end if;
            --  with this statement we transfer the current request
            --+ to the queue of the private channel Assign
            --+ (the transfer is executed directly,
            --+  without evaluating the guard associated to Assign)
            requeue Assign;
         end if;
      end Allocate;
      --+----------
      entry Assign (Res    :    out Resource_T;
                    Amount :        Request_T) when Available is
         --  users will enter this service only if there are
         --+ enough spare resources AND those users had previously
         --+ been refused service
         --+---------
         --  THE ALGORITHM ASSUMES FIFO QUEUEING POLICY!
      begin
         --  one less request in queue
         Considered := Considered - 1;
         --  was it the last one?
         if Considered = 0 then
            --  if it was we close the guard to this channel
            Available := False;
            --  we should now update Min_Request!
         end if;
         --  can we now satisfy this request?
         if Amount <= Free then
            --  yes, this requeued user can now be granted
            --+ what it initially wanted
            Res.Granted := Amount;
            Free := Free - Amount;
         else
            --  otherwise we must requeue the unlucky user
            --+ again and update the record of the smallest
            --+ pending request
            if Amount < Min_Request then
               Min_Request := Amount;
            end if;
            requeue Assign;
         end if;
      end Assign;
      --+----------
      procedure Release (Res    : Resource_T) is
      begin
         --  this service is executed when a user
         --+ returns its allocation of resources
         Free := Free + Res.Granted;
         --  as the level of reserve has now increased
         --+ we should now take a look the requests that
         --+ are awaiting in the Assign queue
         if Free > Min_Request then
            if Assign'Count > 0 then
               --  we do NOT know the user that put forward
               --+ the smallest request, hence, we simply
               --+ take note of the number of requests currently
               --+ enqueued at Assign and open the guard to it
               Considered := Assign'Count;
               Available := True;
               --  and finally we reset the value of Min_Request
               --+ so that it may be recomputed by the execution
               --+ of Assign
               Min_Request := Request_T'Last;
            end if;
         end if;
      end Release;
   end Controller;
   --+----------------------

end RCP.Control;
